from fastapi import BackgroundTasks

from ehelply_microservice_library.integrations.logging import Log
from ehelply_microservice_library.integrations.monitor import Monitor, UsageMQ
from ehelply_microservice_library.utils.ehelply_cloud import (
    ehelply_cloud_access,
    CloudParticipantAuthRequest,
    CloudParticipantResponse,
)


class ManagerBase:
    def __init__(
        self,
            cloud_participant: CloudParticipantResponse | None,
            logger: Log | None, monitor: Monitor | None, background_tasks: BackgroundTasks | None
    ):
        self.cloud_participant: CloudParticipantResponse | None = cloud_participant
        self.logger: Log | None = logger
        self.monitor: Monitor | None = monitor
        self.background_tasks: BackgroundTasks | None = background_tasks

        self.usage: dict[str, int] = {}

    @classmethod
    async def with_auth(
            cls, auth: CloudParticipantAuthRequest,
            logger: Log | None, monitor: Monitor | None, background_tasks: BackgroundTasks | None
    ):
        cloud_participant: CloudParticipantResponse = await ehelply_cloud_access(auth=auth)
        return cls(cloud_participant, logger, monitor, background_tasks)

    def add_usage(self, usage_type: str, amount: int):
        if usage_type in self.usage:
            self.usage[usage_type] += amount
        else:
            self.usage[usage_type] = amount

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.background_tasks and self.monitor and self.cloud_participant:
            for usage_type, amount in self.usage.items():
                self.background_tasks.add_task(self.monitor.add_usage, UsageMQ(
                    project_uuid=self.cloud_participant.project_uuid,
                    usage_key=usage_type,
                    quantity=amount
                ))



from time import sleep
from typing import Union

from fastapi import WebSocket
from fastapi.websockets import WebSocketDisconnect

from ehelply_bootstrapper.sockets.managers import SCMApiGateway, SCMFastApi
from ehelply_bootstrapper.sockets.schemas import ChannelSocketMessage, APIGatewayMessage
from ehelply_bootstrapper.utils.state import State

from ehelply_microservice_library.integrations.m2m import M2M


async def scm_fastapi_init(
        realtime_controller: SCMFastApi,
        websocket: WebSocket,
        m2m: M2M
):
    """
    Call from a FastAPI socket endpoint set up in a microservice router.

    Verifies auth on the initial connection only. Identity token only needs to be valid for the initial connection.

    Args:
        realtime_controller:
        websocket:
        m2m:

    Returns:

    Microservice code example:
    ```
    from ehelply_bootstrapper.sockets.backbones import SCBRedis
    from ehelply_bootstrapper.sockets.managers import SCMFastApi
    from ehelply_bootstrapper.sockets.events import EventSocketEcho, EListenerSocketEcho

    class RealtimeController(SCMFastApi):
        def register_events(self):
            self.register_event(action="ehelply.bootstrapper.socket.echo", event=EventSocketEcho)

        def register_event_listeners(self):
            self.register_listener(EListenerSocketEcho())

    scb_redis: SCBRedis = SCBRedis(State.redis, connection_limit_per_identifier=1)
    realtime_controller: RealtimeController = RealtimeController(scb_redis)

    @router.websocket('/game')
    async def websocket_endpoint(websocket: WebSocket, m2m: M2M = Depends(get_m2m)):
        m2m: M2M = State.integrations.get("m2m")
        scm_fastapi_init(
            realtime_controller=realtime_controller,
            websocket=websocket,
            m2m=m2m
        )
    ```
    """
    authorized: bool = False

    participant_uuid: Union[None, str] = None

    await websocket.accept()

    try:
        while not authorized:
            message: Union[ChannelSocketMessage, dict, None] = await websocket.receive_json(mode='binary')

            if message:
                try:
                    message = ChannelSocketMessage(**message)
                except:
                    message = None

                if message and message.action == "authorization":

                    authorization: str = message.data['authorization']
                    ehelply_active_participant: str = message.data['ehelply-active-participant']

                    try:
                        claims = m2m.verify_token(token=authorization)
                    except:
                        await websocket.close(1008)
                        return

                    if not ehelply_active_participant:
                        await websocket.close(1008)
                        return

                    if ehelply_active_participant in claims['custom:participants'].split(','):
                        participant_uuid = ehelply_active_participant
                        authorized = True
                    else:
                        await websocket.close(1008)
                        return

            else:
                sleep(0.1)

    except WebSocketDisconnect:
        if participant_uuid:
            await realtime_controller.disconnect(participant_uuid)

    if authorized:
        await realtime_controller.connect(participant_uuid, websocket)


async def scm_apigateway_init(
        realtime_controller: SCMApiGateway,
        apigateway_message: APIGatewayMessage,
        m2m: M2M,
        log_messages: bool = False,
        log_errors: bool = False
):
    """
    Call from a realtime HTTP endpoint set up in a microservice router.

    Verifies authorization on every message. Identity token must be valid for each request.

    Excepts messages from API Gateway to have the following form (use Request Templates in APIGateway to achieve this):
    {
        "connection_id": "<APIGateway_ConnectionId>",       # This comes from APIGateway context vars
        "event": "MESSAGE",                                 # This comes from APIGateway context vars
        "message": {                                        # This is the only part that would come from a client
            "action": "<message_action>",
            "data": {
                "authorization": "<identity_token>",
                "ehelply-active-participant": "<active_participant>",
                ... other message data
            }
        }
    }


    Microservice code example:
    ```
    from ehelply_bootstrapper.sockets.schemas import ChannelSocketMessage, APIGatewayMessage
    from ehelply_bootstrapper.sockets.backbones import SCBDict
    from ehelply_bootstrapper.sockets.managers import SCMApiGateway
    from ehelply_bootstrapper.sockets.events import EventSocketEcho, EListenerSocketEcho
    from ehelply_bootstrapper.drivers.aws_utils.aws_apigateway import APIGateway, APIConfig
    from ehelply_bootstrapper.utils.environment import Environment


    class RealtimeController(SCMApiGateway):
        def register_events(self):
            self.register_event(action="ehelply.bootstrapper.socket.echo", event=EventSocketEcho)

        def register_event_listeners(self):
            self.register_listener(EListenerSocketEcho())


    stage: str = Environment.stage()
    if Environment.is_dev():
        stage = "test"

    realtime_backbone: SCBDict = SCBDict()
    realtime_apigateway: APIGateway = APIGateway(
        api_config=APIConfig(
            api_uuid="lt7x6mhfxg",
            stage=stage
        ),
        is_websockets=True
    )
    realtime_controller: RealtimeController = RealtimeController(
        socket_connection_backbone=realtime_backbone,
        api_gateway=realtime_apigateway
    )


    @router.post(
        '/realtime',
        tags=["realtime"],
    )
    async def apigateway_websocket_endpoint(apigateway_message: APIGatewayMessage = Body(...), m2m: M2M = Depends(get_m2m)):
        scm_apigateway_init(realtime_controller=realtime_controller, apigateway_message=apigateway_message, m2m=m2m)
    ```
    """
    # TODO: Auth could and likely should be moved into a lambda function in API Gateway

    if log_messages:
        State.logger.info(str(apigateway_message))

    if apigateway_message.event == "DISCONNECT":
        await realtime_controller.disconnect(identifier=None, connection=apigateway_message.connection_id)
        return

    if apigateway_message.event == "MESSAGE":

        try:
            message: ChannelSocketMessage = apigateway_message.message
        except:
            if log_errors:
                State.logger.warning("eHelply Sockets - Invalid message")
            await realtime_controller.disconnect(identifier=None, connection=apigateway_message.connection_id)
            return

        if 'authorization' not in message.data or 'ehelply-active-participant' not in message.data:
            if log_errors:
                State.logger.warning("eHelply Sockets - Invalid authorization")
            await realtime_controller.disconnect(identifier=None, connection=apigateway_message.connection_id)
            return

        authorization: str = message.data['authorization']
        ehelply_active_participant: str = message.data['ehelply-active-participant']

        try:
            claims = m2m.verify_token(token=authorization)
        except:
            if log_errors:
                State.logger.warning("eHelply Sockets - Invalid authorization token")
            await realtime_controller.disconnect(identifier=None, connection=apigateway_message.connection_id)
            return

        if not ehelply_active_participant:
            if log_errors:
                State.logger.warning("eHelply Sockets - Invalid active participant")
            await realtime_controller.disconnect(identifier=None, connection=apigateway_message.connection_id)
            return

        if ehelply_active_participant in claims['custom:participants'].split(','):
            participant_uuid = ehelply_active_participant
        else:
            if log_errors:
                State.logger.warning("eHelply Sockets - Active participant is not allowed")
            await realtime_controller.disconnect(identifier=None, connection=apigateway_message.connection_id)
            return

        del message.data['authorization']
        del message.data['ehelply-active-participant']

        await realtime_controller.connect(participant_uuid, apigateway_message.connection_id)
        await realtime_controller.on_receive(ehelply_active_participant, message)



from ehelply_bootstrapper.integrations.integration import Integration
from ehelply_bootstrapper.utils.state import State
from ehelply_microservice_library.integrations.fact import get_fact_endpoint
from pydantic import BaseModel
import json
from fastapi.encoders import jsonable_encoder
from typing import Union, List, Dict, Any, Tuple


class NotificationRelation(BaseModel):
    service: Union[str, None] = None
    product: Union[str, None] = None
    type: Union[str, None] = None
    target: Union[str, None] = None
    link: Union[str, None] = None


class NotificationContent(BaseModel):
    raw: str  # Plain text string. ie. Broadview clinic is delayed by 5 minutes.
    html: Union[
        str, None] = None  # HTML formatted string. ie. <a href="sdfgndgh">Broadview clinic</a> is <strong>delayed</strong> by 5 minutes.


class Notification(BaseModel):
    subject: str
    content: NotificationContent
    relations: List[
        NotificationRelation] = []  # If there is only a single relation, clicking the notification in a UI should take you directly to the relation. If there are multiple, it should take you to a notifications page.
    severity: str = "low"  # low, med, high
    type: str  # ie. ehelply-appointments.appointment.delayed <- used to determine where a notification gets published to
    picture: Union[str, None] = None


class NotificationType(BaseModel):
    email_template: str = "ehelply-generic"
    publisher_types: List[str] = [
        "database"
    ]
    provider_whitelist: Dict[str, List[str]] = {
        "database": [
            "dynamo"
        ]
    }
    type: str


class Notifications(Integration):
    def __init__(self, notification_types: List[NotificationType]) -> None:
        """
        Setup the microservice
        """
        super().__init__("notifications")

        # M2M contains a configured python requests instance which you should use for API requests.
        #  ie. self.m2m.requests
        self.m2m = State.integrations.get("m2m")

        self.notification_types: List[NotificationType] = notification_types

        self.sqs = None

    def init(self):
        """
        STEP 0: Initialize the integration.
        :return:
        """
        super().init()

    def load(self):
        """
        STEP 1: Load the integration
        :return:
        """
        self.sqs = State.aws.make_client("sqs")

        try:
            for notification_type in self.notification_types:
                self.create_type(notification_type)

        except:
            State.logger.warning("Unable to create notification types.")

    @staticmethod
    def get_base_url() -> str:
        """
        Get the base URL of this microservice
        :return:
        """
        return get_fact_endpoint('ehelply-notifications')

    def create_type(self, notification_type: NotificationType):
        response = self.m2m.requests.post(self.get_base_url() + "/types", json={"type": notification_type.dict()})

        return response.json()

    def push_notification(self, notification: Notification):
        queue_url: str = get_fact_endpoint("mq-notifications")
        self.sqs.send_message(
            QueueUrl=queue_url,
            MessageBody=json.dumps(jsonable_encoder(notification.dict()))
        )
